"""Base model for 1D Gaussian mixture model (GMM)."""

import math
from typing import Any, Optional

import torch
import torch.distributions as dist
import torch.nn.functional as F
from torch import nn

from flowgym.registry import base_model_registry
from flowgym.schedulers import OptimalTransportScheduler, Scheduler
from flowgym.types import FlowTensor
from flowgym.utils import append_dims, train_base_model

from .base import BaseModel


@base_model_registry.register("1d/gmm")
class OneDimensionalBaseModel(BaseModel[FlowTensor]):
    """Base model for 1D Gaussian mixture model (GMM).

    Keep in mind that this trains the model, so it may take a minute to load.
    """

    output_type = "velocity"

    def __init__(
        self,
        device: Optional[torch.device],
        scheduler: Optional[Scheduler[FlowTensor]] = None,
    ):
        super().__init__(device)

        if device is None:
            device = torch.device("cpu")

        self.device = device

        if scheduler is None:
            scheduler = OptimalTransportScheduler()

        self._scheduler = scheduler

        p1 = dist.MixtureSameFamily(  # type: ignore
            dist.Categorical(torch.ones(2)),  # type: ignore
            dist.Normal(torch.Tensor([0.0, 3.0]), torch.Tensor([1.0, 0.4])),  # type: ignore
        )
        data = [FlowTensor(p1.sample((4096, 1)).to(device))]  # type: ignore

        mlp = MLP(1, 1).to(device)
        opt = torch.optim.Adam(mlp.parameters(), lr=1e-3)

        train_base_model(self, data, epochs=1_000, batch_size=512, opt=opt, pbar=True)

    @property
    def scheduler(self) -> Scheduler[FlowTensor]:
        """Optimal transport scheduler."""
        return self._scheduler

    def sample_p0(self, n: int, **kwargs: Any) -> tuple[FlowTensor, dict[str, Any]]:
        """Sample n data points from the base distribution p0.

        Parameters
        ----------
        n : int
            Number of samples to draw.

        kwargs : dict
            Additional keyword arguments (unused).

        Returns
        -------
        samples : FlowTensor, shape (n, 1)
            Samples from the base distribution p0.

        kwargs : dict
            Additional keyword arguments (unused).
        """
        return FlowTensor(torch.randn(n, 1, device=self.device)), kwargs

    def forward(self, x: FlowTensor, t: torch.Tensor, **kwargs: Any) -> FlowTensor:
        """Forward pass of the base model.

        Parameters
        ----------
        x : FlowTensor, shape (n, 1)
            Input data.

        t : torch.Tensor, shape (n,)
            Time steps, values in [0, 1].

        Returns
        -------
        output : FlowTensor
            Output of the model.
        """
        return FlowTensor(self.model(x.data, t))


class MLP(nn.Module):
    """Multi-layer perceptron."""

    def __init__(
        self,
        in_dim: int,
        out_dim: int,
        time_dim: int = 128,
        cond_dim: int = 256,
        depth: int = 3,
        width: int = 256,
        window_size: float = 1000.0,
        t_mult: float = 1000.0,
    ) -> None:
        super().__init__()

        self.time_embed = SinusoidalTimeEmbedding(time_dim, cond_dim, window_size, t_mult)

        blocks = [Block(in_dim, width, cond_dim)]
        for _ in range(depth - 1):
            blocks.append(Block(width, width, cond_dim))

        self.blocks = nn.ModuleList(blocks)
        self.head = nn.Linear(width, out_dim, bias=True)

        nn.init.zeros_(self.head.weight)
        nn.init.zeros_(self.head.bias)

    def forward(self, x: torch.Tensor, t: torch.Tensor) -> torch.Tensor:
        """Forward pass of the MLP.

        Parameters
        ----------
        x : torch.Tensor, shape (n, input_dim)
            Input data.

        t : torch.Tensor, shape (n,)
            Time steps, values in [0, 1].

        Returns
        -------
        output : torch.Tensor, shape (n, output_dim)
            Output of the MLP.
        """
        cond = self.time_embed(t)
        x_ = x.flatten(start_dim=t.ndim)
        for block in self.blocks:
            x_ = block(x_, cond)

        return self.head(x_).reshape(*x.shape[:-1], -1)


class Block(nn.Module):
    """A single block of the MLP."""

    def __init__(self, in_dim: int, out_dim: int, cond_dim: int):
        super().__init__()
        self.linear = nn.Linear(in_dim, out_dim)
        self.norm = nn.LayerNorm(out_dim)
        self.film = FiLM(out_dim, cond_dim)

    def forward(self, x: torch.Tensor, cond: torch.Tensor) -> torch.Tensor:
        """Forward pass of the block."""
        return F.silu(self.film(self.norm(self.linear(x)), cond))


class SinusoidalTimeEmbedding(nn.Module):
    """t -> R^d sinusoidal embedding + MLP."""

    def __init__(
        self,
        dim: int = 128,
        hidden_dim: int = 256,
        window_size: float = 1000.0,
        t_mult: float = 1000.0,
    ) -> None:
        super().__init__()
        self.mlp = nn.Sequential(
            nn.Linear(dim, hidden_dim), nn.SiLU(), nn.Linear(hidden_dim, hidden_dim)
        )

        self.t_mult = t_mult
        half = dim // 2
        freqs = torch.exp(-math.log(window_size) * torch.arange(half, dtype=torch.float32) / half)
        self.register_buffer("freqs", freqs)

    def forward(self, t: torch.Tensor) -> torch.Tensor:
        """Compute sinusoidal time embedding."""
        # Assuming t is in [0, 1], scale to [0, t_mult] as is usual for diffusion models
        t = t.float() * self.t_mult
        args = t.unsqueeze(-1) * self.freqs.unsqueeze(-2)
        emb = torch.cat([torch.sin(args), torch.cos(args)], dim=-1)
        result: torch.Tensor = self.mlp(emb)
        return result


class FiLM(nn.Module):
    """Feature-wise linear modulation h -> gamma(t) * h + beta(t)."""

    def __init__(self, n_channels: int, cond_dim: int):
        super().__init__()
        self.to_scale_shift = nn.Linear(cond_dim, 2 * n_channels)

    def forward(self, x: torch.Tensor, cond: torch.Tensor) -> torch.Tensor:
        """Apply FiLM modulation."""
        gamma, beta = self.to_scale_shift(cond).chunk(2, dim=-1)
        gamma = append_dims(gamma, x.ndim)
        beta = append_dims(beta, x.ndim)
        return x * (1 + gamma) + beta



"""Abstract base class for base models used in flow matching and diffusion."""

from abc import ABC, abstractmethod
from typing import Any, Generic, Literal, Optional

import torch
from torch import nn

from flowgym.schedulers import Scheduler
from flowgym.types import D

OutputType = Literal["epsilon", "endpoint", "velocity", "score"]


class BaseModel(ABC, nn.Module, Generic[D]):
    """Abstract base class for base models used in flow matching and diffusion."""

    output_type: OutputType

    def __init__(self, device: Optional[torch.device]):
        super().__init__()

        if device is None:
            device = torch.device("cpu")

        self.device = device

    @property
    @abstractmethod
    def scheduler(self) -> Scheduler[D]:
        """Base model-dependent scheduler used for sampling."""

    @abstractmethod
    def sample_p0(self, n: int, **kwargs: Any) -> tuple[D, dict[str, Any]]:
        """Sample n data points from the base distribution p0.

        Parameters
        ----------
        n : int
            Number of samples to draw.

        **kwargs : dict
            Additional keyword arguments.

        Returns
        -------
        samples : D
            Samples from the base distribution p0.

        kwargs : dict
            Additional keyword arguments.
        """

    @abstractmethod
    def forward(self, x: D, t: torch.Tensor, **kwargs: Any) -> D:
        """Forward pass of the base model.

        Parameters
        ----------
        x : D
            Input data.

        t : torch.Tensor, shape (n,)
            Time steps, values in [0, 1].

        Returns
        -------
        output : D
            Output of the model.
        """

    def preprocess(self, x: D, **kwargs: Any) -> tuple[D, dict[str, Any]]:
        """Preprocess data and keyword arguments for the base model.

        Parameters
        ----------
        x : D
            Input data to preprocess.

        **kwargs : dict
            Additional keyword arguments to preprocess.

        Returns
        -------
        output : D
            Preprocessed data.

        kwargs : dict
            Preprocessed keyword arguments.
        """
        return x, kwargs

    def postprocess(self, x: D) -> D:
        """Postprocess samples x_1 (e.g., decode with VAE).

        Parameters
        ----------
        x : D
            Input data to postprocess.

        Returns
        -------
        output : D
            Postprocessed output.
        """
        return x



"""Pre-trained continuous FlowMol model, trained on GEOM-Drugs."""

from typing import Any

import dgl
import flowmol
import torch
from flowmol.data_processing.utils import build_edge_idxs, get_batch_idxs, get_upper_edge_mask

from flowgym import BaseModel, ConstantNoiseSchedule, CosineScheduler, Scheduler
from flowgym.molecules.types import FlowGraph
from flowgym.registry import base_model_registry
from flowgym.types import FlowTensor


class FlowMolBaseModel(BaseModel[FlowGraph]):
    """Pre-trained FlowMol on GEOM-Drugs and QM9."""

    output_type = "endpoint"

    def __init__(
        self,
        model_name: str,
        scheduler_params: tuple[float, float, float, float],
        device: torch.device,
    ):
        super().__init__(device)

        self.model = flowmol.load_pretrained(model_name).to(device)
        self._scheduler = FlowMolScheduler(scheduler_params)
        self._scheduler.noise_schedule = ConstantNoiseSchedule(0)

    @property
    def scheduler(self) -> "FlowMolScheduler":
        """Scheduler used for sampling."""
        return self._scheduler

    def sample_p0(self, n: int, **kwargs: Any) -> tuple[FlowGraph, dict[str, Any]]:
        """Sample n datapoints from the base distribution :math:`p_0`.

        Parameters
        ----------
        n : int
            Number of samples to draw.

        Returns
        -------
        samples : FlowGraph
            Samples from the base distribution :math:`p_0`.

        Notes
        -----
        The base distribution :math:`p_0` is a standard Gaussian distribution.
        """
        n_atoms = self.model.sample_n_atoms(n)

        edge_idxs_dict = {}
        for n_atoms_i in torch.unique(n_atoms):
            edge_idxs_dict[int(n_atoms_i)] = build_edge_idxs(n_atoms_i)

        g = []
        for n_atoms_i in n_atoms:
            edge_idxs = edge_idxs_dict[int(n_atoms_i)]
            g_i = dgl.graph((edge_idxs[0], edge_idxs[1]), num_nodes=n_atoms_i, device=self.device)
            g.append(g_i)

        g = dgl.batch(g)
        ue_mask = get_upper_edge_mask(g)
        n_idx, e_idx = get_batch_idxs(g)
        g = self.model.sample_prior(g, n_idx, ue_mask)

        # Set all x_0 to x_t (this is what FlowMol expects)
        for key in list(g.ndata.keys()):
            if key.endswith("_0"):
                g.ndata[key[:-2] + "_t"] = g.ndata.pop(key)

        for key in list(g.edata.keys()):
            if key.endswith("_0"):
                g.edata[key[:-2] + "_t"] = g.edata.pop(key)

        return FlowGraph(g, ue_mask, n_idx, e_idx), kwargs

    def postprocess(self, x: FlowGraph) -> FlowGraph:
        """Re-name features from x_t to x_1."""
        g = x.graph
        for key in list(g.ndata.keys()):
            if key.endswith("_t"):
                g.ndata[key[:-2] + "_1"] = g.ndata.pop(key)

        for key in list(g.edata.keys()):
            if key.endswith("_t"):
                g.edata[key[:-2] + "_1"] = g.edata.pop(key)

        # To enable usage with SampledMolecule from flowmol
        g.edata["ue_mask"] = x.ue_mask
        return FlowGraph(g, x.ue_mask, x.n_idx, x.e_idx)

    def forward(self, x: FlowGraph, t: torch.Tensor, **kwargs: Any) -> FlowGraph:
        r"""Compute the endpoint vector field :math:`\hat{x_1}(x, t)`."""
        output = self.model.vector_field(
            x.graph,
            t,
            x.n_idx,
            x.ue_mask,
            apply_softmax=True,
            remove_com=True,
        )

        out_graph = x._get_empty_graph()
        for key in x.graph.ndata:
            out_graph.ndata[key] = output[key[:-2]]

        for key in x.graph.edata:
            data = x.graph.edata[key]
            if isinstance(data, torch.Tensor):
                # Output only contains upper edge data, need to expand to both upper and lower
                edge_data = torch.zeros_like(data)
                edge_data[x.ue_mask] = output[key[:-2]]  # Assign to upper edges
                edge_data[~x.ue_mask] = output[key[:-2]]  # Assign same values to lower edges
                out_graph.edata[key] = edge_data

        return FlowGraph(out_graph, x.ue_mask, x.n_idx, x.e_idx)


class FlowMolScheduler(Scheduler[FlowGraph]):
    """Scheduler for the GEOM-Drugs dataset."""

    def __init__(
        self,
        cosine_params: tuple[float, float, float, float] = (1.0, 2.0, 2.0, 2.0),
    ) -> None:
        self.schedulers = {
            "x_t": CosineScheduler(cosine_params[0]),
            "a_t": CosineScheduler(cosine_params[1]),
            "c_t": CosineScheduler(cosine_params[2]),
            "e_t": CosineScheduler(cosine_params[3]),
        }
        self.scheduler_order = ["x_t", "a_t", "c_t", "e_t"]

    def _alpha(self, x: FlowGraph, t: torch.Tensor) -> torch.Tensor:
        out = torch.zeros(t.shape[0], len(self.schedulers), device=t.device, dtype=t.dtype)
        for idx, key in enumerate(self.scheduler_order):
            val = self.schedulers[key].alpha(
                FlowTensor(torch.zeros(x.graph.batch_size, 1, device=x.device)), t
            )
            out[:, idx] = val.data.squeeze(-1)

        return out

    def alpha(self, x: FlowGraph, t: torch.Tensor) -> FlowGraph:
        r""":math:`\alpha_t`."""
        g = x.graph
        res = x._get_empty_graph()

        alphas = self._alpha(x, t)
        for idx, key in enumerate(self.scheduler_order):
            if key in g.ndata:
                t_alpha = alphas[:, idx].unsqueeze(-1)
                res.ndata[key] = t_alpha[x.n_idx]
            elif key in g.edata:
                t_alpha = alphas[:, idx].unsqueeze(-1)
                res.edata[key] = t_alpha[x.e_idx]
            else:
                raise ValueError(f"Key {key} not found in graph data.")

        return FlowGraph(res, x.ue_mask, x.n_idx, x.e_idx)

    def _alpha_dot(self, x: FlowGraph, t: torch.Tensor) -> torch.Tensor:
        out = torch.zeros(t.shape[0], len(self.schedulers), device=t.device, dtype=t.dtype)
        for idx, key in enumerate(self.scheduler_order):
            val = self.schedulers[key].alpha_dot(
                FlowTensor(torch.zeros(x.graph.batch_size, 1, device=x.device)), t
            )
            out[:, idx] = val.data.squeeze(-1)

        return out

    def alpha_dot(self, x: FlowGraph, t: torch.Tensor) -> FlowGraph:
        r""":math:`\dot{\alpha}_t`."""
        g = x.graph
        res = x._get_empty_graph()

        alpha_dots = self._alpha_dot(x, t)
        for idx, key in enumerate(self.scheduler_order):
            if key in g.ndata:
                t_alpha_dot = alpha_dots[:, idx].unsqueeze(-1)
                res.ndata[key] = t_alpha_dot[x.n_idx]
            elif key in g.edata:
                t_alpha_dot = alpha_dots[:, idx].unsqueeze(-1)
                res.edata[key] = t_alpha_dot[x.e_idx]
            else:
                raise ValueError(f"Key {key} not found in graph data.")

        return FlowGraph(res, x.ue_mask, x.n_idx, x.e_idx)


@base_model_registry.register("molecules/flowmol_geom")
class GEOMBaseModel(FlowMolBaseModel):
    """Pre-trained continuous flow matching model, trained on GEOM-Drugs."""

    def __init__(self, device: torch.device):
        super().__init__("geom_gaussian", (1.0, 2.0, 2.0, 2.0), device)


@base_model_registry.register("molecules/flowmol_qm9")
class QM9BaseModel(FlowMolBaseModel):
    """Pre-trained continuous flow matching model, trained on QM9."""

    def __init__(self, device: torch.device):
        super().__init__("qm9_gaussian", (1.0, 2.0, 2.0, 1.5), device)



"""Common schedulers for flow matching and diffusion models."""

from typing import cast

import torch

from flowgym.types import FlowTensor
from flowgym.utils import append_dims

from .base import Scheduler


class OptimalTransportScheduler(Scheduler[FlowTensor]):
    r"""Optimal transport scheduler which is commonly used to train flow matching models.

    Schedule:
    .. math::

        \alpha_t = t, \quad \beta_t = 1 - t, \quad \dot{\alpha}_t = 1, \quad \dot{\beta}_t = -1.
    """

    def alpha(self, x: FlowTensor, t: torch.Tensor) -> FlowTensor:
        r""":math:`\alpha_t = t`."""
        return FlowTensor(append_dims(t, x.data.ndim))

    def alpha_dot(self, x: FlowTensor, t: torch.Tensor) -> FlowTensor:
        r""":math:`\dot{\alpha}_t = 1`."""
        return x.ones_like()


class CosineScheduler(Scheduler[FlowTensor]):
    """Cosine scheduler."""

    def __init__(self, nu: float):
        self.nu = nu

    def alpha(self, x: FlowTensor, t: torch.Tensor) -> FlowTensor:
        alpha = 1 - torch.cos((t**self.nu) * torch.pi / 2).square()
        return FlowTensor(append_dims(alpha, x.data.ndim))

    def alpha_dot(self, x: FlowTensor, t: torch.Tensor) -> FlowTensor:
        alpha_dot = (
            self.nu
            * torch.pi
            * (t ** (self.nu - 1))
            * torch.sin((t**self.nu) * torch.pi / 2)
            * torch.cos((t**self.nu) * torch.pi / 2)
        )
        return FlowTensor(append_dims(alpha_dot, x.data.ndim))


class DiffusionScheduler(Scheduler[FlowTensor]):
    """Scheduler for discrete-time diffusion models based on a given noise schedule.

    Parameters
    ----------
    alpha_bar : torch.Tensor
        Cumulative product of (1 - beta) values, shape (K,), where K is the number of diffusion
        steps.
    """

    def __init__(self, alpha_bar: torch.Tensor):
        super().__init__()

        self.alpha_bar = alpha_bar
        self.alpha_bar_shifted = torch.cat(
            [torch.ones(1, device=alpha_bar.device, dtype=alpha_bar.dtype), alpha_bar[:-1]], dim=0
        )
        self.K = alpha_bar.shape[0] - 1
        self.alpha_bar_dot = self.K * (self.alpha_bar_shifted - self.alpha_bar)

    def _get_index(self, t: torch.Tensor) -> torch.Tensor:
        k = ((1 - t) * self.K + 0.5).long().clamp(0, self.K).cpu()
        return cast("torch.Tensor", k)

    def model_input(self, t: torch.Tensor) -> torch.Tensor:
        """Input to the model at time t that encodes the timestep."""
        return self._get_index(t).to(t.device)

    def alpha(self, x: FlowTensor, t: torch.Tensor) -> FlowTensor:
        k = self._get_index(t)
        alpha = torch.sqrt(self.alpha_bar[k])
        return FlowTensor(append_dims(alpha, x.data.ndim))

    def beta(self, x: FlowTensor, t: torch.Tensor) -> FlowTensor:
        k = self._get_index(t)
        beta = torch.sqrt(1 - self.alpha_bar[k])
        return FlowTensor(append_dims(beta, x.data.ndim))

    def alpha_dot(self, x: FlowTensor, t: torch.Tensor) -> FlowTensor:
        k = self._get_index(t)
        alpha = torch.sqrt(self.alpha_bar[k])
        alpha_dot = 0.5 * self.alpha_bar_dot[k] / alpha
        return FlowTensor(append_dims(alpha_dot, x.data.ndim))

    def beta_dot(self, x: FlowTensor, t: torch.Tensor) -> FlowTensor:
        k = self._get_index(t)
        beta = torch.sqrt(1 - self.alpha_bar[k])
        beta_dot = -0.5 * self.alpha_bar_dot[k] / beta
        return FlowTensor(append_dims(beta_dot, x.data.ndim))



"""Custom types for flowgym."""

from typing import Callable, Protocol, Sequence, TypeVar, Union

import torch
from typing_extensions import Self

UnaryOp = Callable[[torch.Tensor], torch.Tensor]
BinaryOp = Callable[[torch.Tensor, torch.Tensor], torch.Tensor]
D = TypeVar("D", bound="FlowMixin")


class FlowProtocol(Protocol):
    def __len__(self) -> int: ...
    def __getitem__(self, idx: Union[int, slice]) -> Self: ...

    @classmethod
    def collate(cls: type[Self], items: Sequence[Self]) -> Self: ...

    def aggregate(self) -> torch.Tensor:
        """Reduce over all dimensions except batch (i.e., sum per sample).

        Returns: Tensor of shape (len(self),)
        """
        ...

    def apply(self, op: UnaryOp) -> Self:
        """Apply a function to the underlying tensor data (e.g., x -> -x)."""
        ...

    def combine(self, other: Self, op: BinaryOp) -> Self:
        """Combine self with another instance element-wise (e.g., x + y).

        Note: `other` is guaranteed to be of the same type as `self` because the Mixin handles
        scalar broadcasting before calling this.
        """
        ...


class FlowMixin(FlowProtocol):
    """Mixin for common functionality."""

    @property
    def device(self) -> torch.device:
        # Find the device of the underlying data and make sure all underlying data is on the same
        # device
        dev = None

        def get_tensor(x: torch.Tensor) -> torch.Tensor:
            nonlocal dev
            if dev is None:
                dev = x.device

            if dev != x.device:
                raise RuntimeError(f"Inconsistent devices found in {type(self)}.")

            return x

        self.apply(get_tensor)

        if dev is None:
            raise RuntimeError(f"No tensors found in {type(self)} to determine device.")

        return dev

    def to(self, device: torch.device | str) -> Self:
        return self.apply(lambda x: x.to(device))

    def _binary_dispatch(self, other: Union[Self, float, torch.Tensor], op: BinaryOp) -> Self:
        if isinstance(other, torch.Tensor):
            return self.apply(lambda x: op(x, other))

        if isinstance(other, (int, float)):
            return self.apply(lambda x: op(x, torch.tensor(other, device=x.device, dtype=x.dtype)))

        if type(other) is type(self):
            return self.combine(other, op)

        raise TypeError(
            f"Unsupported operand type(s) for operation: {type(self)} and {type(other)}"
        )

    def __add__(self, other):
        return self._binary_dispatch(other, torch.add)

    def __sub__(self, other):
        return self._binary_dispatch(other, torch.sub)

    def __mul__(self, other):
        return self._binary_dispatch(other, torch.mul)

    def __truediv__(self, other):
        return self._binary_dispatch(other, torch.div)

    def __neg__(self):
        return self.apply(torch.neg)

    def __pow__(self, power):
        return self.apply(lambda x: torch.pow(x, power))

    def __radd__(self, other):
        return self._binary_dispatch(other, lambda x, y: torch.add(y, x))

    def __rsub__(self, other):
        return self._binary_dispatch(other, lambda x, y: torch.sub(y, x))

    def __rmul__(self, other):
        return self._binary_dispatch(other, lambda x, y: torch.mul(y, x))

    def randn_like(self) -> Self:
        return self.apply(torch.randn_like)

    def ones_like(self) -> Self:
        return self.apply(torch.ones_like)

    def zeros_like(self) -> Self:
        return self.apply(torch.zeros_like)

    def clone(self) -> Self:
        return self.apply(torch.clone)

    def requires_grad(self, mode: bool = True) -> Self:
        """In-place operation to start recording operations for autograd."""
        return self.apply(lambda x: x.requires_grad_(mode))

    def gradient(
        self, outputs: torch.Tensor, create_graph: bool = False, retain_graph: bool = False
    ) -> Self:
        """Compute the gradient of output w.r.t. self.

        Returns: An instance of Self containing the gradients.
        """
        # Gather inputs
        inputs = []

        def gather_inputs(x: torch.Tensor) -> torch.Tensor:
            inputs.append(x)
            return x

        self.apply(gather_inputs)

        # Compute gradients
        grad_outputs = torch.ones_like(outputs) if outputs.ndim > 0 else None
        raw_grads = torch.autograd.grad(
            outputs=outputs,
            inputs=inputs,
            grad_outputs=grad_outputs,
            create_graph=create_graph,
            retain_graph=retain_graph,
            allow_unused=True,
        )

        # Inject gradients (this is in the same order as we gathered the inputs)
        grad_iter = iter(raw_grads)

        def inject_grads(x: torch.Tensor) -> torch.Tensor:
            g = next(grad_iter)
            return g if g is not None else torch.zeros_like(x)

        return self.apply(inject_grads)


class FlowTensor(FlowMixin):
    """A FlowType wrapper around torch.Tensor."""

    def __init__(self, data: torch.Tensor):
        if not isinstance(data, torch.Tensor):
            raise TypeError("FlowTensor expects a torch.Tensor")

        if data.ndim < 1:
            raise ValueError("FlowTensor expects a tensor with at least 1 dimension")

        self.data = data

    def __repr__(self) -> str:
        return (
            f"{type(self).__name__}("
            f"shape={tuple(self.data.shape)}, "
            f"dtype={self.data.dtype}, "
            f"device={self.data.device})"
        )

    def __len__(self) -> int:
        return self.data.shape[0]

    def __getitem__(self, idx: Union[int, slice]) -> Self:
        data_out = self.data[idx]

        if data_out.ndim < self.data.ndim:
            data_out = data_out.unsqueeze(0)

        return type(self)(data_out)

    @classmethod
    def collate(cls, items: Sequence[Self]) -> Self:
        if not items:
            raise ValueError("Cannot collate an empty sequence")

        tensors = [item.data for item in items]
        return cls(torch.cat(tensors, dim=0))

    def aggregate(self) -> torch.Tensor:
        return self.data.sum(dim=tuple(range(1, self.data.ndim)))

    def apply(self, op: UnaryOp) -> Self:
        return type(self)(op(self.data))

    def combine(self, other: Self, op: BinaryOp) -> Self:
        return type(self)(op(self.data, other.data))



"""Common noise schedules for flow matching and diffusion models."""

import torch

from flowgym.types import D

from .base import NoiseSchedule


class ConstantNoiseSchedule(NoiseSchedule[D]):
    """Constant noise schedule with fixed sigma.

    Parameters
    ----------
    sigma : float
        Constant noise level.
    """

    def __init__(self, sigma: float):
        self.sigma = sigma

    def __call__(self, x: D, t: torch.Tensor) -> D:
        """Constant noise schedule."""
        return self.sigma * x.ones_like()



"""Compression-based reward implementations."""

import io
from typing import Any

import torch
from torchvision.transforms.functional import to_pil_image

from flowgym import FlowTensor, Reward
from flowgym.registry import reward_registry


def _bits_per_pixel(imgs: torch.Tensor, quality_level: int) -> torch.Tensor:
    IMG_BATCH_NDIM = 4
    assert imgs.ndim == IMG_BATCH_NDIM, "imgs should be a batch of images with shape (B, C, H, W)"

    if imgs.min() < 0 or imgs.max() > 1:
        raise ValueError(f"`imgs` must have values in [0, 1], got [{imgs.min()}, {imgs.max()}]")

    batch_size = imgs.shape[0]
    pixels = imgs.shape[-1] * imgs.shape[-2]

    bpp = torch.zeros(batch_size, device=imgs.device)
    for i in range(batch_size):
        # Convert to PIL Image
        pil_image = to_pil_image(imgs[i].cpu().float())

        # Calculate compressed size (using JPEG) in bytes
        compressed_buffer = io.BytesIO()
        pil_image.save(compressed_buffer, format="JPEG", quality=quality_level, optimize=True)
        compressed_size = len(compressed_buffer.getvalue())

        # Convert to bits/pixel
        bpp[i] = compressed_size * 8 / pixels

    return bpp


@reward_registry.register("images/incompression")
class IncompressionReward(Reward[FlowTensor]):
    """Incompression reward for image models.

    Typically, when this reward is maximized, it encourages the model to produce images that have
    high detail and patterns.

    Parameters
    ----------
    quality_level : int, 1-100
        JPEG quality level. Lower values mean higher compression.

    Examples
    --------
    >>> reward = IncompressionReward(85)
    >>> imgs = torch.rand(8, 3, 256, 256)
    >>> result = reward(imgs)
    >>> result.shape
    torch.Size([8])
    """

    def __init__(self, quality_level: int = 85):
        self.quality_level = quality_level

    def __call__(self, x: FlowTensor, **kwargs: Any) -> tuple[torch.Tensor, torch.Tensor]:
        """Compute the incompression reward for a batch of images.

        Parameters
        ----------
        imgs : tensor, shape (B, C, H, W), values in [0, 1]
            A batch of images.

        Returns
        -------
        rewards : torch.Tensor, shape (B,)
            Incompression reward (bits per pixel) for each image.
        """
        return _bits_per_pixel(x.data, self.quality_level), torch.ones(len(x), device=x.device)


@reward_registry.register("images/compression")
class CompressionReward(Reward[FlowTensor]):
    """Compression reward for image models.

    Typically, when this reward is maximized, it encourages the model to produce images that look
    more vintage or like paintings.

    Parameters
    ----------
    quality_level : int, 1-100
        JPEG quality level. Lower values mean higher compression.

    Examples
    --------
    >>> reward = CompressionReward(85)
    >>> imgs = torch.rand(8, 3, 256, 256)
    >>> result = reward(imgs)
    >>> result.shape
    torch.Size([8])
    """

    def __init__(self, quality_level: int = 85):
        self.quality_level = quality_level

    def __call__(self, x: FlowTensor, **kwargs: Any) -> tuple[torch.Tensor, torch.Tensor]:
        """Compute the compression reward for a batch of images.

        Parameters
        ----------
        x : tensor, shape (B, C, H, W), values in [0, 1]
            A batch of images.

        Returns
        -------
        rewards : torch.Tensor, shape (B,)
            Compression reward (negative bits per pixel) for each image.
        """
        return -_bits_per_pixel(x.data, self.quality_level), torch.ones(len(x), device=x.device)



"""Types for molecular graphs in Flow Gym."""

from typing import Any, Optional, Sequence, Union

import dgl
import torch
from torch_scatter import scatter
from typing_extensions import Self

from flowgym.types import BinaryOp, FlowMixin, UnaryOp


def construct_ue_mask(g: dgl.DGLGraph) -> torch.Tensor:
    """Construct a mask indicating upper edges in the graph."""
    ul_pattern = torch.tensor([1, 0], device=g.device).repeat(g.batch_size)
    n_edges_pattern = (g.batch_num_edges() / 2).int().repeat_interleave(2)
    return ul_pattern.repeat_interleave(n_edges_pattern).bool()


def construct_n_idx(g: dgl.DGLGraph) -> torch.Tensor:
    """Construct a tensor which maps each node to its graph index in the batch."""
    return torch.repeat_interleave(
        torch.arange(g.batch_size, device=g.device),
        g.batch_num_nodes(),
    )


def construct_e_idx(g: dgl.DGLGraph) -> torch.Tensor:
    """Construct a tensor which maps each edge to its graph index in the batch."""
    return torch.repeat_interleave(
        torch.arange(g.batch_size, device=g.device),
        g.batch_num_edges(),
    )


class FlowGraph(FlowMixin):
    """A wrapper around DGLGraph that supports required factory methods.

    Parameters
    ----------
    graph : dgl.DGLGraph
        The graph to wrap.

    ue_mask : Optional[torch.Tensor], optional
        Mask indicating upper edges in the graph, by default None

    n_idx : Optional[torch.Tensor], optional
        Tensor mapping each node to its graph index in the batch, by default None

    e_idx : Optional[torch.Tensor], optional
        Tensor mapping each edge to its graph index in the batch, by default None
    """

    def __init__(
        self,
        graph: dgl.DGLGraph,
        ue_mask: Optional[torch.Tensor] = None,
        n_idx: Optional[torch.Tensor] = None,
        e_idx: Optional[torch.Tensor] = None,
    ):
        if ue_mask is None:
            ue_mask = construct_ue_mask(graph)
        if n_idx is None:
            n_idx = construct_n_idx(graph)
        if e_idx is None:
            e_idx = construct_e_idx(graph)

        self.graph = graph
        self.ue_mask = ue_mask
        self.n_idx = n_idx
        self.e_idx = e_idx

    def __repr__(self) -> str:
        return (
            f"{type(self).__name__}("
            f"num_nodes={self.graph.num_nodes()}, "
            f"num_edges={self.graph.num_edges()}, "
            f"batch_size={len(self)})"
        )

    def __getattr__(self, name: str) -> Any:
        return getattr(self.graph, name)

    def __setattr__(self, name: str, value: Any) -> None:
        if name == "graph":
            object.__setattr__(self, name, value)
        else:
            setattr(self.graph, name, value)

    def __len__(self) -> int:
        return int(self.graph.batch_size)

    def __getitem__(self, idx: Union[int, slice]) -> Self:
        if isinstance(idx, int):
            # Faster to use slice_batch if we only want one item
            if idx < 0:
                idx += len(self)

            if idx < 0 or idx >= len(self):
                raise IndexError(f"Index {idx} out of range for batch size {len(self)}")

            return type(self)(dgl.slice_batch(self.graph, idx))

        if isinstance(idx, slice):
            graphs = dgl.unbatch(self.graph)
            selected_graphs = graphs[idx]

            if not selected_graphs:
                raise ValueError("The slice resulted in an empty graph sequence.")

            return type(self)(dgl.batch(selected_graphs))

        raise TypeError(f"Invalid index type: {type(idx)}")

    @classmethod
    def collate(cls, items: Sequence[Self]) -> Self:
        if not items:
            raise ValueError("Cannot collate an empty sequence")

        return cls(dgl.batch([item.graph for item in items]))

    def _get_empty_graph(self) -> dgl.DGLGraph:
        """Get an empty graph with the same structure as Self."""
        # Clone the graph structure
        empty_graph = dgl.graph(self.graph.edges(), num_nodes=self.graph.num_nodes())

        # Preserve batch information
        if self.graph.batch_size > 1:
            empty_graph.set_batch_num_nodes(self.graph.batch_num_nodes())
            empty_graph.set_batch_num_edges(self.graph.batch_num_edges())

        return empty_graph

    def apply(self, op: UnaryOp) -> Self:
        res = self._get_empty_graph()

        for key, val in self.graph.ndata.items():
            if isinstance(val, torch.Tensor):
                res.ndata[key] = op(val)

        for key, val in self.graph.edata.items():
            if isinstance(val, torch.Tensor):
                res.edata[key] = op(val)

        return type(self)(res, self.ue_mask, self.n_idx, self.e_idx)

    def combine(self, other: Union[Self, float, torch.Tensor], op: BinaryOp) -> Self:
        res = self._get_empty_graph()

        if isinstance(other, FlowGraph):
            for key, val in self.graph.ndata.items():
                if key in other.graph.ndata:
                    res.ndata[key] = op(val, other.graph.ndata[key])  # type: ignore
                else:
                    res.ndata[key] = val

            for key, val in self.graph.edata.items():
                if key in other.graph.edata:
                    res.edata[key] = op(val, other.graph.edata[key])  # type: ignore
                else:
                    res.edata[key] = val
        else:
            for key, val in self.graph.ndata.items():
                res.ndata[key] = op(val, other)  # type: ignore

            for key, val in self.graph.edata.items():
                res.edata[key] = op(val, other)  # type: ignore

        return type(self)(res, self.ue_mask, self.n_idx, self.e_idx)

    def aggregate(self) -> torch.Tensor:
        summed = torch.zeros(self.graph.batch_size, device=self.graph.device)
        for _, val in self.graph.ndata.items():
            if isinstance(val, torch.Tensor):
                summed += scatter(val, self.n_idx, dim=0, reduce="sum").sum(dim=-1)

        for _, val in self.graph.edata.items():
            if isinstance(val, torch.Tensor):
                summed += scatter(val, self.e_idx, dim=0, reduce="sum").sum(dim=-1)

        return summed



"""Dipole moment reward for molecules."""

import json
import os
from json.decoder import JSONDecodeError
from typing import Any, Optional

import numpy as np
import torch
from numpy.typing import NDArray
from rdkit import Chem

from flowgym.molecules.rewards.base import BatchedMoleculeReward
from flowgym.registry import reward_registry
from flowgym.utils import temporary_workdir


@reward_registry.register("molecules/dipole_moment")
class DipoleMomentReward(BatchedMoleculeReward):
    """Dipole moment reward for molecules. Requires xtb to be installed."""

    def compute_rewards(
        self,
        mols: list[Chem.Mol],
        indices: list[int],
        **kwargs: Any,
    ) -> tuple[torch.Tensor, torch.Tensor]:
        """Run GFN2-xTB to compute the dipole moments of the molecules in Debye."""
        results = parallel_xtb(mols, gfn2_opt=True)
        rewards = torch.tensor(
            [float(np.linalg.norm(res.dipole)) if res is not None else 0.0 for res in results],
            dtype=torch.float32,
        )
        rewards = rewards * 2.5417  # Convert from atomic units to Debye
        valids = torch.tensor([res is not None for res in results])

        return rewards, valids


@reward_registry.register("molecules/target_dipole_moment")
class TargetDipoleMomentReward(DipoleMomentReward):
    """MAE to a target dipole moment for molecules."""

    invalid_reward = -2.0

    def compute_rewards(
        self,
        mols: list[Chem.Mol],
        indices: list[int],
        **kwargs: Any,
    ) -> tuple[torch.Tensor, torch.Tensor]:
        """Absolute error to dipole moment target."""
        target: Optional[torch.Tensor] = kwargs.get("dipole_moment_target", None)
        if target is None:
            raise ValueError(
                "Target dipole moment must be provided "
                "as a keyword argument 'dipole_moment_target'."
            )

        target = target[indices]
        if target.shape[0] != len(mols):
            raise ValueError(
                f"Target dipole moment ({target.shape[0]}) must have the same length "
                f"as the number of molecules ({len(mols)})."
            )

        dipoles, valids = super().compute_rewards(mols, indices, **kwargs)
        rewards = -torch.abs(dipoles - target)

        return rewards, torch.tensor(valids, dtype=torch.bool)


class XTBResult:
    """Class to parse the output of GFN2-xTB."""

    def __init__(self, filename: str):
        assert filename.endswith(".json"), f"Filename ({filename}) must end with .json"
        with open(filename, "r") as f:
            self.json_data = json.load(f)

    @property
    def energy(self) -> float:
        """Energy in Hartree."""
        return float(self.json_data["total energy"])

    @property
    def dipole(self) -> NDArray[np.float32]:
        """Dipole in atomic units."""
        for key, value in self.json_data.items():
            if key.startswith("dipole"):
                return np.array(value)

        raise RuntimeError(
            "Dipole moment not found in xtb output. "
            f"Keys in output: {sorted(self.json_data.keys())}"
        )


def parallel_xtb(mols: list[Chem.Mol], gfn2_opt: bool = False) -> list[XTBResult | None]:
    """Run GFN2-xTB in parallel for a the molecules in the graph."""
    results = []
    with temporary_workdir():
        i = 0
        for mol in mols:
            i += 1
            Chem.MolToXYZFile(mol, f"{i}.xyz")

        ncpus = len(os.sched_getaffinity(0))

        # Compute properties using GFN2-xTB
        opt_str = "--opt" if gfn2_opt else ""
        os.system(
            f"parallel -j {ncpus} "
            f"'xtb {{}} {opt_str} --parallel 1 --namespace {{/.}} --json > {{/.}}.out 2>&1' "
            "::: *.xyz"
        )

        # Read results
        for i in range(1, len(mols) + 1):
            path = f"{i}.xtbout.json"

            try:
                res = XTBResult(path) if os.path.exists(path) else None
            except JSONDecodeError:
                res = None

            results.append(res)

    return results



"""Base reward classes and interfaces for flowgym."""

from abc import ABC, abstractmethod
from typing import Any, Generic

import torch

from flowgym.types import D


class Reward(ABC, Generic[D]):
    """Abstract base class for all rewards."""

    @abstractmethod
    def __call__(self, x: D, **kwargs: Any) -> tuple[torch.Tensor, torch.Tensor]:
        """Compute the reward and validity for the given input x."""



"""Schedulers for flow matching and diffusion models."""

from .base import MemorylessNoiseSchedule, NoiseSchedule, Scheduler
from .noise_schedules import ConstantNoiseSchedule
from .schedulers import CosineScheduler, DiffusionScheduler, OptimalTransportScheduler

__all__ = [
    "ConstantNoiseSchedule",
    "CosineScheduler",
    "DiffusionScheduler",
    "MemorylessNoiseSchedule",
    "NoiseSchedule",
    "OptimalTransportScheduler",
    "Scheduler",
]



"""Pre-trained base model for Stable Diffusion 2."""

import json
import random
from typing import TYPE_CHECKING, Any, Optional, cast

import torch
from diffusers.pipelines.stable_diffusion.pipeline_stable_diffusion import StableDiffusionPipeline

from flowgym import BaseModel, DiffusionScheduler, FlowTensor
from flowgym.registry import base_model_registry

if TYPE_CHECKING:
    from diffusers.models.unets.unet_2d import UNet2DModel


class StableDiffusionBaseModel(BaseModel[FlowTensor]):
    """Stable Diffusion base model."""

    output_type = "epsilon"

    def __init__(self, model_name: str, device: Optional[torch.device]):
        super().__init__(device)

        pipe = StableDiffusionPipeline.from_pretrained(model_name).to(device)
        self.pipe = pipe
        self.unet: UNet2DModel = pipe.unet

        self.channels: int = self.unet.config["in_channels"]
        self.dim: int = self.unet.config["sample_size"]

        pipe.scheduler.alphas_cumprod = pipe.scheduler.alphas_cumprod.to(device)
        self._scheduler = DiffusionScheduler(pipe.scheduler.alphas_cumprod)

        with open("flowgym/images/base_models/refl_data.json", "r") as f:
            refl_data = json.load(f)

        self.prompts = [item["text"] for item in refl_data]

    @property
    def scheduler(self) -> DiffusionScheduler:
        """Scheduler used for sampling."""
        return self._scheduler

    def sample_p0(self, n: int, **kwargs: Any) -> tuple[FlowTensor, dict[str, Any]]:
        """Sample n latent datapoints from the base distribution :math:`p_0`.

        Parameters
        ----------
        n : int
            Number of samples to draw.

        kwargs : dict
            Additional keyword arguments.

        Returns
        -------
        samples : FlowTensor, shape (n, 4, 64, 64)
            Samples from the base distribution :math:`p_0`.

        kwargs : dict
            Additional keyword arguments, a randomly selected prompt if not provided through the
            input.

        Notes
        -----
        The base distribution :math:`p_0` is a standard Gaussian distribution.
        """
        prompt = kwargs.get("prompt", None)
        cfg_scale = kwargs.get("cfg_scale", 0.0)

        # If no prompt is provided, sample them
        if prompt is None:
            prompt = random.sample(self.prompts, n)

        # If a single prompt is provided, replicate it
        if not isinstance(prompt, list):
            prompt = [prompt] * n

        if len(prompt) != n:
            raise ValueError(
                "The prompt must be a list of strings with length equal to the batch "
                f"size, got length {len(prompt)}."
            )

        return (
            FlowTensor(torch.randn(n, self.channels, self.dim, self.dim, device=self.device)),
            {"prompt": prompt, "cfg_scale": cfg_scale},
        )

    def preprocess(self, x: FlowTensor, **kwargs: Any) -> tuple[FlowTensor, dict[str, Any]]:
        """Encode the prompt (if provided instead of encoder_hidden_states).

        Parameters
        ----------
        x : FlowTensor, shape (n, 4, 64, 64)
            Input data to preprocess.

        **kwargs : dict
            Additional keyword arguments to preprocess.

        Returns
        -------
        output : DataType
            Preprocessed data.

        kwargs : dict
            Preprocessed keyword arguments.
        """
        n = len(x)
        encoder_hidden_states = kwargs.get("encoder_hidden_states")
        prompt = kwargs.get("prompt")
        neg_prompt = kwargs.get("neg_prompt", "")
        cfg_scale = kwargs.get("cfg_scale", 0.0)

        if encoder_hidden_states is None and prompt is None:
            raise ValueError("Either encoder_hidden_states or a prompt needs to be provided.")

        if encoder_hidden_states is None:
            if not isinstance(prompt, list):
                prompt = [prompt] * n

            assert len(prompt) == n, (
                "The prompt must be a list of strings with length equal to the batch size."
            )

            if neg_prompt is not None:
                if not isinstance(neg_prompt, list):
                    neg_prompt = [neg_prompt] * n

                if len(neg_prompt) != n:
                    raise ValueError(
                        "The negative prompt must be a list of strings with length equal to the ",
                        "batch size.",
                    )

            prompt_embeds, neg_prompt_embeds = self.pipe.encode_prompt(
                prompt, self.device, 1, cfg_scale > 0, neg_prompt
            )
            encoder_hidden_states = prompt_embeds
            if neg_prompt_embeds is not None:
                encoder_hidden_states = torch.cat([prompt_embeds, neg_prompt_embeds], dim=0)
            else:
                encoder_hidden_states = prompt_embeds

        return x, {
            "encoder_hidden_states": encoder_hidden_states,
            "cfg_scale": cfg_scale,
            "prompt": prompt,
            "neg_prompt": neg_prompt,
        }

    def postprocess(self, x: FlowTensor) -> FlowTensor:
        """Decode the images from the latent space.

        Parameters
        ----------
        x : FlowTensor, shape (n, 4, 64, 64)
            Final sample in latent space.

        Returns
        -------
        decoded : FlowTensor, shape (n, 3, 512, 512)
            Decoded images in pixel space.
        """
        # Do this one-by-one to save on a lot of VRAM
        x = x / self.pipe.vae.config.scaling_factor
        decoded = torch.cat([self.pipe.vae.decode(xi.unsqueeze(0)).sample for xi in x.data], dim=0)

        # Convert to [0, 1]
        decoded = (decoded + 1) / 2
        decoded = decoded.clamp(0, 1)

        return FlowTensor(decoded)

    def forward(self, x: FlowTensor, t: torch.Tensor, **kwargs: Any) -> FlowTensor:
        r"""Forward pass of the model, outputting :math:`\epsilon(x_t, t)`.

        Parameters
        ----------
        x : FlowTensor, shape (n, 4, 64, 64)
            Input data.

        t : torch.Tensor, shape (n,)
            Time steps, values in [0, 1].

        **kwargs : dict
            Additional keyword arguments passed to the UNet model.

        Returns
        -------
        output : FlowTensor, shape (n, 4, 64, 64)
            Output of the model.
        """
        x_tensor = x.data
        k = self.scheduler.model_input(t)

        cfg_scale = kwargs.get("cfg_scale", 0.0)
        encoder_hidden_states = kwargs.get("encoder_hidden_states")
        if encoder_hidden_states is None:
            raise ValueError("encoder_hidden_states must be provided in kwargs.")

        if isinstance(cfg_scale, torch.Tensor):
            use_cfg = torch.any(cfg_scale > 0).item()
        else:
            use_cfg = cfg_scale > 0

        if not use_cfg:
            out = cast("torch.Tensor", self.unet(x_tensor, k, encoder_hidden_states).sample)
            return FlowTensor(out)

        x_tensor = torch.cat([x_tensor, x_tensor], dim=0)
        k = torch.cat([k, k], dim=0)
        out = cast("torch.Tensor", self.unet(x_tensor, k, encoder_hidden_states).sample)

        # Classifier-free guidance
        cond, uncond = out.chunk(2)
        out = (cfg_scale + 1) * cond - cfg_scale * uncond

        return FlowTensor(out)


@base_model_registry.register("images/sd2")
class SD2BaseModel(StableDiffusionBaseModel):
    """Pre-trained 512x512 Stable Diffusion 2 base."""

    def __init__(self, device: Optional[torch.device]):
        super().__init__("PeggyWang/stable-diffusion-2-base", device)


@base_model_registry.register("images/sd1.5")
class SD15BaseModel(StableDiffusionBaseModel):
    """Pre-trained 512x512 Stable Diffusion 1.5."""

    def __init__(self, device: Optional[torch.device]):
        super().__init__("sd-legacy/stable-diffusion-v1-5", device)



"""Utility functions for flowgym."""

from __future__ import annotations

import os
import tempfile
from contextlib import contextmanager
from typing import TYPE_CHECKING, Any, Generator, Generic

import torch
from torch import nn
from torch.utils.data import DataLoader, Dataset
from tqdm import tqdm
from flowgym.schedulers import NoiseSchedule
from flowgym.types import D

if TYPE_CHECKING:
    from flowgym.base_models import BaseModel


def append_dims(x: torch.Tensor, ndim: int) -> torch.Tensor:
    """Match the number of dimensions of x to ndim by adding dimensions at the end.

    Parameters
    ----------
    x : torch.Tensor, shape (*shape)
        The input tensor.

    ndim : int
        The target number of dimensions.

    Returns
    -------
    x : torch.Tensor, shape (*shape, 1, ..., 1)
        The reshaped tensor with ndim dimensions.
    """
    if x.ndim > ndim:
        return x

    shape = x.shape + (1,) * (ndim - x.ndim)
    return x.view(shape)


@contextmanager
def temporary_workdir() -> Generator[str, None, None]:
    """Context manager that runs code in a fresh temporary directory.

    When exiting the context, it returns to the original working directory and deletes the temporary
    folder.
    """
    old_cwd = os.getcwd()
    with tempfile.TemporaryDirectory() as tmp:
        try:
            os.chdir(tmp)
            yield tmp
        finally:
            os.chdir(old_cwd)


class ValuePolicy(nn.Module, Generic[D]):
    r"""Policy based on a value function, :math:`u(x, t) = -\sigma(t) \nabla_x V(x, t)`.

    Parameters
    ----------
    value_network : nn.Module
        The value function network, :math:`V(x, t)`.

    noise_schedule : NoiseSchedule
        The noise schedule, :math:`\sigma(t)`.
    """

    def __init__(self, value_network: nn.Module, noise_schedule: NoiseSchedule[D]) -> None:
        super().__init__()
        self.value_network = value_network
        self.noise_schedule = noise_schedule

    @torch.enable_grad()  # type: ignore[no-untyped-call]
    def forward(self, x: D, t: torch.Tensor, **kwargs: Any) -> D:
        """Compute control action based on value function gradient."""
        x = x.requires_grad()
        value_pred = self.value_network(x, t, **kwargs)
        sigma = self.noise_schedule(x, t)
        control: D = -sigma * x.gradient(value_pred)
        return control


class FlowDataset(Dataset[D]):
    """Dataset wrapper for flowgym data."""

    def __init__(self, data: list[D]):
        if len(data) == 0:
            raise ValueError("Data list is empty.")

        # Combine all data into a single object
        self.data = type(data[0]).collate(data)

    def __len__(self) -> int:
        return len(self.data)

    def __getitem__(self, idx: int) -> D:
        return self.data[idx]


def train_base_model(
    base_model: BaseModel[D],
    data: list[D],
    epochs: int,
    batch_size: int,
    opt: torch.optim.Optimizer,
    pbar: bool = False,
) -> None:
    """Trains/fine-tunes a base model.

    Parameters
    ----------
    model : BaseModel[D]
        The model to train.

    data : list[D]
        The training data.

    epochs : int
        The number of training epochs.
    """
    base_model.train()
    scheduler = base_model.scheduler

    dataset = FlowDataset(data)
    loader = DataLoader(dataset, batch_size, shuffle=True, collate_fn=type(data[0]).collate)

    iterator = range(epochs)
    if pbar:
        iterator = tqdm(iterator)

    for _ in iterator:
        total_loss = 0.0
        for batch in loader:
            x1 = batch[0]
            x1: D

            x1 = x1.to(base_model.device)
            x0 = x1.randn_like()
            t = torch.rand(len(x1), device=x1.device)

            alpha = scheduler.alpha(x1, t)
            beta = scheduler.beta(x1, t)

            xt = alpha * x1 + beta * x0
            pred = base_model(xt, t)

            if base_model.output_type == "velocity":
                alpha_dot = scheduler.alpha_dot(x1, t)
                beta_dot = scheduler.beta_dot(x1, t)
                target = alpha_dot * x1 + beta_dot * x0
            elif base_model.output_type == "endpoint":
                target = x1
            elif base_model.output_type == "epsilon":
                target = x0
            elif base_model.output_type == "score":
                target = -x0 / beta
            else:
                raise ValueError(f"Unknown output type: {base_model.output_type}")

            loss = ((pred - target) ** 2).aggregate().mean()

            opt.zero_grad()
            loss.backward()
            opt.step()

            total_loss += loss.item() * len(x1)

        if isinstance(iterator, tqdm):
            iterator.set_postfix({"loss": total_loss / len(dataset)})



"""Pre-trained base model for Diffusion Transformer."""

from typing import Any, Optional, cast

import torch
from diffusers.pipelines.dit.pipeline_dit import DiTPipeline

from flowgym import BaseModel, DiffusionScheduler, FlowTensor
from flowgym.registry import base_model_registry


@base_model_registry.register("images/dit")
class DiTBaseModel(BaseModel[FlowTensor]):
    """Pre-trained 256x256 ImageNet transformer diffusion model.

    Uses the `facebookresearch/DiT-XL-2-256` model from the `diffusers` library.
    """

    output_type = "epsilon"

    def __init__(self, device: Optional[torch.device]):
        super().__init__(device)

        pipe = DiTPipeline.from_pretrained(
            "facebook/DiT-XL-2-256",
        ).to(device)
        self.pipe = pipe
        self.transformer = pipe.transformer

        self.channels: int = self.transformer.config["in_channels"]
        self.dim: int = self.transformer.config["sample_size"]

        pipe.scheduler.alphas_cumprod = pipe.scheduler.alphas_cumprod.to(device)
        self._scheduler = DiffusionScheduler(pipe.scheduler.alphas_cumprod)

    @property
    def scheduler(self) -> DiffusionScheduler:
        """Scheduler used for sampling."""
        return self._scheduler

    def sample_p0(self, n: int, **kwargs: Any) -> tuple[FlowTensor, dict[str, Any]]:
        """Sample n latent datapoints from the base distribution :math:`p_0`.

        Parameters
        ----------
        n : int
            Number of samples to draw.

        kwargs : dict
            Additional keyword arguments.

        Returns
        -------
        samples : FlowTensor, shape (n, 4, 64, 64)
            Samples from the base distribution :math:`p_0`.

        kwargs : dict
            Additional keyword arguments, a randomly selected class label is provided if
            "class_label" is not in the input.

        Notes
        -----
        The base distribution :math:`p_0` is a standard Gaussian distribution.
        """
        class_labels = kwargs.get("class_labels", None)
        cfg_scale = kwargs.get("cfg_scale", 0.0)

        # If no prompt is provided, sample them
        if class_labels is None:
            class_labels = torch.randint(0, 1000, (n,), device=self.device)

        if isinstance(class_labels, int):
            class_labels = torch.tensor([class_labels] * n, device=self.device)

        if isinstance(class_labels, list):
            class_labels = torch.tensor(class_labels, device=self.device)

        if len(class_labels) != n:
            raise ValueError(
                "The class_label must be a list of integers with length equal to the "
                f"batch size, got length {len(class_labels)}."
            )

        return (
            FlowTensor(torch.randn(n, self.channels, self.dim, self.dim, device=self.device)),
            {"class_labels": class_labels, "cfg_scale": cfg_scale},
        )

    def preprocess(self, x: FlowTensor, **kwargs: Any) -> tuple[FlowTensor, dict[str, Any]]:
        """Encode the prompt (if provided instead of encoder_hidden_states).

        Parameters
        ----------
        x : FlowTensor, shape (n, 4, 64, 64)
            Input data to preprocess.

        **kwargs : dict
            Additional keyword arguments to preprocess.

        Returns
        -------
        output : DataType
            Preprocessed data.

        kwargs : dict
            Preprocessed keyword arguments.
        """
        class_labels = kwargs.get("class_labels", None)
        cfg_scale = kwargs.get("cfg_scale", 0.0)

        if class_labels is None:
            raise ValueError("class_labels must be provided in kwargs.")

        return x, {
            "cfg_scale": cfg_scale,
            "class_labels": class_labels,
        }

    def postprocess(self, x: FlowTensor) -> FlowTensor:
        """Decode the images from the latent space.

        Parameters
        ----------
        x : FlowTensor, shape (n, 4, 64, 64)
            Final sample in latent space.

        Returns
        -------
        decoded : FlowTensor, shape (n, 3, 512, 512)
            Decoded images in pixel space.
        """
        # Do this one-by-one to save on a lot of VRAM
        x = x / self.pipe.vae.config.scaling_factor
        decoded = torch.cat([self.pipe.vae.decode(xi.unsqueeze(0)).sample for xi in x.data], dim=0)

        # Convert to [0, 1]
        decoded = (decoded + 1) / 2
        decoded = decoded.clamp(0, 1)

        return FlowTensor(decoded)

    def forward(self, x: FlowTensor, t: torch.Tensor, **kwargs: Any) -> FlowTensor:
        r"""Forward pass of the model, outputting :math:`\epsilon(x_t, t)`.

        Parameters
        ----------
        x : FlowTensor, shape (n, 4, 64, 64)
            Input data.

        t : torch.Tensor, shape (n,)
            Time steps, values in [0, 1].

        **kwargs : dict
            Additional keyword arguments passed to the UNet model.

        Returns
        -------
        output : FlowTensor, shape (n, 4, 64, 64)
            Output of the model.
        """
        x_tensor = x.data
        k = self.scheduler.model_input(t)

        cfg_scale = kwargs.get("cfg_scale", 0.0)
        class_labels = kwargs.get("class_labels")

        if class_labels is None:
            raise ValueError("class_labels must be provided in kwargs.")

        if isinstance(cfg_scale, torch.Tensor):
            use_cfg = torch.any(cfg_scale > 0).item()
        else:
            use_cfg = cfg_scale > 0

        if not use_cfg:
            out = cast("torch.Tensor", self.transformer(x_tensor, k, class_labels).sample[:, :4])
            return FlowTensor(out)

        x_tensor = torch.cat([x_tensor, x_tensor], dim=0)
        k = torch.cat([k, k], dim=0)
        class_null = torch.zeros_like(class_labels)
        class_labels = torch.cat([class_labels, class_null], dim=0)
        out = cast("torch.Tensor", self.transformer(x_tensor, k, class_labels).sample)

        # Classifier-free guidance
        cond, uncond = out[:, :4].chunk(2)
        out = (cfg_scale + 1) * cond - cfg_scale * uncond

        return FlowTensor(out)



"""Base classes for schedulers of flow matching models."""

from abc import ABC, abstractmethod
from typing import Generic

import torch

from flowgym.types import D


class NoiseSchedule(ABC, Generic[D]):
    """Abstract base class for noise schedules."""

    @abstractmethod
    def __call__(self, x: D, t: torch.Tensor) -> D:
        r"""Compute the noise level at time t.

        Can be overwritten if the noise schedule is data-dependent.

        Parameters
        ----------
        x : D
            Data tensor.

        t : torch.Tensor, shape (n,)
            Time tensor with values in [0, 1].

        Returns
        -------
        sigma_t : D, same data shape as x
            Values of :math:`\sigma(t)` at the given times.
        """
        ...


class Scheduler(ABC, Generic[D]):
    r"""Abstract base class for schedulers of flow matching models.

    Generally :math:`\beta_t = 1-\alpha_t`, but this can be re-defined. Furthermore, generally we
    are interested in a memoryless noise schedule, which is the default of `noise_schedule` (i.e.,
    :math:`\sigma`), however this can also be re-defined.
    """

    @property
    def noise_schedule(self) -> NoiseSchedule[D]:
        """Get the current noise schedule."""
        if not hasattr(self, "_noise_schedule"):
            self._noise_schedule: NoiseSchedule[D] = MemorylessNoiseSchedule(self)

        return self._noise_schedule

    @noise_schedule.setter
    def noise_schedule(self, schedule: NoiseSchedule[D]) -> None:
        """Set the noise schedule. Defaults to the memoryless noise schedule."""
        self._noise_schedule = schedule

    @abstractmethod
    def alpha(self, x: D, t: torch.Tensor) -> D:
        r""":math:`\alpha_t`.

        Can be overwritten if :math:`\alpha_t` is data-dependent.

        Parameters
        ----------
        x : D
            Data tensor.

        t : torch.Tensor, shape (n,)
            Time tensor with values in [0, 1].

        Returns
        -------
        alpha_t : D, same data shape as x
            Values of :math:`\alpha_t` at the given times.
        """
        ...

    @abstractmethod
    def alpha_dot(self, x: D, t: torch.Tensor) -> D:
        r""":math:`\dot{\alpha}_t`.

        Can be overwritten if :math:`\dot{\alpha}_t` is data-dependent.

        Parameters
        ----------
        x : D
            Data tensor.

        t : torch.Tensor, shape (n,)
            Time tensor with values in [0, 1].

        Returns
        -------
        alpha_dot_t : D, same data shape as x
            Values of :math:`\dot{\alpha}_t` at the given times.
        """
        ...

    def beta(self, x: D, t: torch.Tensor) -> D:
        r""":math:`\beta_t`.

        Parameters
        ----------
        x : D
            Data tensor.

        t : torch.Tensor, shape (n,)
            Time tensor with values in [0, 1].

        Returns
        -------
        beta_t : D, same data shape as x
            Values of :math:`\beta_t` at the given times.
        """
        return 1 - self.alpha(x, t)

    def beta_dot(self, x: D, t: torch.Tensor) -> D:
        r""":math:`\dot{\beta}_t`.

        Can be overwritten if :math:`\dot{\beta}_t` is data-dependent.

        Parameters
        ----------
        x : D
            Data tensor.

        t : torch.Tensor, shape (n,)
            Time tensor with values in [0, 1].

        Returns
        -------
        beta_dot_t : D, same data shape as x
            Values of :math:`\dot{\beta}_t` at the given times.
        """
        return -self.alpha_dot(x, t)

    def sigma(self, x: D, t: torch.Tensor) -> D:
        r""":math:`\sigma(t)` noise schedule.

        Can be overwritten if :math:`\sigma(t)` is data-dependent.

        Parameters
        ----------
        x : D
            Data tensor.

        t : torch.Tensor, shape (n,)
            Time tensor with values in [0, 1].

        Returns
        -------
        sigma_t : D, same data shape as x
            Values of :math:`\sigma(t)` at the given times.
        """
        return self.noise_schedule(x, t)

    def model_input(self, t: torch.Tensor) -> torch.Tensor:
        """Input to the model at time t.

        Defaults to t, but could be different if using a different time parameterization.
        """
        return t

    def kappa(self, x: D, t: torch.Tensor) -> D:
        r""":math:`\kappa_t` as defined in [Adjoint Matching](https://openreview.net/forum?id=xQBRrtQM8u)."""
        return self.alpha_dot(x, t) / self.alpha(x, t)

    def eta(self, x: D, t: torch.Tensor) -> D:
        r""":math:`\eta_t` as defined in [Adjoint Matching](https://openreview.net/forum?id=xQBRrtQM8u)."""
        alpha = self.alpha(x, t)
        alpha_dot = self.alpha_dot(x, t)
        beta = self.beta(x, t)
        beta_dot = self.beta_dot(x, t)
        return beta * ((alpha_dot / alpha) * beta - beta_dot)


class MemorylessNoiseSchedule(NoiseSchedule[D]):
    r"""Memoryless noise schedule based on the scheduler's eta function.

    This schedule ensures that :math:`x_0` and :math:`x_1` are independent, which is necessary for
    unbiased generative optimization.

    Parameters
    ----------
    scheduler : Scheduler
        Scheduler to use for computing :math:`\eta_t`.
    """

    def __init__(self, scheduler: Scheduler[D]):
        self.scheduler = scheduler

    def __call__(self, x: D, t: torch.Tensor) -> D:
        r"""Compute the noise level at time t.

        This is given by :math:`\sigma(t) = \sqrt{2 \eta(t)}`.

        Parameters
        ----------
        x : D
            Data tensor.

        t : torch.Tensor, shape (n,)
            Time tensor with values in [0, 1].

        Returns
        -------
        sigma_t : D, same data shape as x
            Values of :math:`\sigma(t)` at the given times.
        """
        return (2 * self.scheduler.eta(x, t)) ** 0.5



"""Optional molecule base models and rewards for Flow Gym."""

from .base_models.flowmol_model import (
    FlowMolBaseModel,
    FlowMolScheduler,
    GEOMBaseModel,
    QM9BaseModel,
)
from .rewards.base import BatchedMoleculeReward, MoleculeReward
from .rewards.dipole_moment import DipoleMomentReward, TargetDipoleMomentReward
from .rewards.protein_docking import ProteinDockingReward
from .rewards.qed import QEDReward
from .rewards.utils import non_fragmented, relax_geometry, validate_mol
from .rewards.validity import ValidityReward
from .types import FlowGraph

__all__ = [
    "BatchedMoleculeReward",
    "DipoleMomentReward",
    "FlowGraph",
    "FlowMolBaseModel",
    "FlowMolScheduler",
    "GEOMBaseModel",
    "MoleculeReward",
    "ProteinDockingReward",
    "QEDReward",
    "QM9BaseModel",
    "TargetDipoleMomentReward",
    "ValidityReward",
    "non_fragmented",
    "relax_geometry",
    "validate_mol",
]



"""Base class for molecule rewards with basic functionalities."""

from abc import abstractmethod
from typing import Any, Optional

import dgl
import torch
from flowmol.analysis.molecule_builder import SampledMolecule
from rdkit import Chem, RDLogger

from flowgym import Reward
from flowgym.molecules.rewards.utils import non_fragmented, relax_geometry, validate_mol
from flowgym.molecules.types import FlowGraph


class MoleculeReward(Reward[FlowGraph]):
    """Base class for molecule rewards with basic functionalities.

    Parameters
    ----------
    atom_type_map : list[str], default=["C", "H", "N", "O", "F", "P", "S", "Cl", "Br", "I"]
        List of atom types corresponding to the node features in the input graph. The index of each
        atom type in the list corresponds to its integer representation in the node features.

    validate : bool, default=False
        Whether to validate molecules using RDKit's sanitization.

    check_fragmented : bool, default=True
        Whether to allow fragmented molecules (i.e., molecules with multiple disconnected
        components).

    relax : bool, default=True
        Whether to relax the geometry of molecules using MMFF optimization. If using `relax`, you
        probably should also use `validate`, since otherwise there is a high chance that your
        program will crash.
    """

    def __init__(
        self,
        atom_type_map: Optional[list[str]] = None,
        validate: bool = False,
        check_fragmented: bool = True,
        relax: bool = True,
    ) -> None:
        RDLogger.DisableLog("rdApp.*")

        if atom_type_map is None:
            atom_type_map = ["C", "H", "N", "O", "F", "P", "S", "Cl", "Br", "I"]

        self.atom_type_map = atom_type_map
        self.validate = validate
        self.check_fragmented = check_fragmented
        self.relax = relax

    def preprocess(self, mol: Chem.Mol) -> Optional[Chem.Mol]:
        """Preprocess the molecule by validation, fragmentation checking, and geometry relaxation.

        Parameters
        ----------
        mol : Chem.Mol
            The molecule to preprocess.

        Returns
        -------
        preprocessed_mol : Chem.Mol | None
            The preprocessed molecule if it passes all checks, else None.
        """
        if self.validate:
            mol = validate_mol(mol)  # type: ignore
            if mol is None:
                return None

        if self.check_fragmented:
            mol = non_fragmented(mol)  # type: ignore
            if mol is None:
                return None

        mol.UpdatePropertyCache(strict=False)

        if self.relax:
            return relax_geometry(mol)

        return mol

    @abstractmethod
    def compute_reward(self, mol: Chem.Mol, **kwargs: Any) -> float | None:
        """Compute the reward for the molecule.

        Parameters
        ----------
        mol : Chem.Mol
            The molecule to compute the reward for.

        Returns
        -------
        reward : float or None
            The computed reward.
        """
        ...

    def __call__(self, x: FlowGraph, **kwargs: Any) -> tuple[torch.Tensor, torch.Tensor]:
        """Compute the reward for the molecule.

        Parameters
        ----------
        mol : FlowGraph
            The molecules to compute the reward for.

        Returns
        -------
        reward : torch.Tensor
            The computed reward.
        """
        i = -1
        out = torch.zeros(x.graph.batch_size, device=x.graph.device)
        valids = torch.zeros(x.graph.batch_size, device=x.graph.device, dtype=torch.bool)
        for g in dgl.unbatch(x.graph):
            i += 1
            mol = SampledMolecule(g.cpu(), self.atom_type_map).rdkit_mol
            if not isinstance(mol, Chem.Mol):
                valids[i] = False
                continue

            mol = self.preprocess(mol)
            if mol is None:
                valids[i] = False
                continue

            output = self.compute_reward(mol, **kwargs)
            if output is None:
                out[i] = 0
                valids[i] = False
            else:
                out[i] = output
                valids[i] = True

        return out, valids


class BatchedMoleculeReward(Reward[FlowGraph]):
    """Base class for molecule rewards with basic functionalities.

    Parameters
    ----------
    atom_type_map : list[str], default=["C", "H", "N", "O", "F", "P", "S", "Cl", "Br", "I"]
        List of atom types corresponding to the node features in the input graph. The index of each
        atom type in the list corresponds to its integer representation in the node features.

    validate : bool, default=False
        Whether to validate molecules using RDKit's sanitization.

    check_fragmented : bool, default=True
        Whether to allow fragmented molecules (i.e., molecules with multiple disconnected
        components).

    relax : bool, default=True
        Whether to relax the geometry of molecules using MMFF optimization. If using `relax`, you
        probably should also use `validate`, since otherwise there is a high chance that your
        program will crash.
    """

    invalid_reward: float = 0.0

    def __init__(
        self,
        atom_type_map: Optional[list[str]] = None,
        validate: bool = False,
        check_fragmented: bool = True,
        relax: bool = False,
    ) -> None:
        RDLogger.DisableLog("rdApp.*")

        if atom_type_map is None:
            atom_type_map = ["C", "H", "N", "O", "F", "P", "S", "Cl", "Br", "I"]

        self.atom_type_map = atom_type_map
        self.validate = validate
        self.check_fragmented = check_fragmented
        self.relax = relax

    def preprocess(self, mol: Chem.Mol) -> Chem.Mol | None:
        """Preprocess the molecule by validation, fragmentation checking, and geometry relaxation.

        Parameters
        ----------
        mol : Chem.Mol
            The molecule to preprocess.

        Returns
        -------
        preprocessed_mol : Chem.Mol | None
            The preprocessed molecule if it passes all checks, else None.
        """
        if self.validate:
            mol = validate_mol(mol)  # type: ignore
            if mol is None:
                return None

        if self.check_fragmented:
            mol = non_fragmented(mol)  # type: ignore
            if mol is None:
                return None

        if self.relax:
            return relax_geometry(mol)

        return mol

    @abstractmethod
    def compute_rewards(
        self,
        mols: list[Chem.Mol],
        indices: list[int],
        **kwargs: Any,
    ) -> tuple[torch.Tensor, torch.Tensor]:
        """Compute the reward for the molecules.

        Parameters
        ----------
        mols : list[Chem.Mol]
            The molecules to compute the reward for.

        Returns
        -------
        reward : torch.Tensor, shape (len(mols),)
            The computed reward.
        """
        ...

    def __call__(self, x: FlowGraph, **kwargs: Any) -> tuple[torch.Tensor, torch.Tensor]:
        """Compute the reward for the molecule.

        Parameters
        ----------
        mol : FlowGraph
            The molecules to compute the reward for.

        Returns
        -------
        reward : torch.Tensor
            The computed reward.
        """
        i = -1
        mols: list[Chem.Mol] = []
        indices = []
        for g in dgl.unbatch(x.graph):
            i += 1
            mol = SampledMolecule(g.cpu(), self.atom_type_map).rdkit_mol
            if not isinstance(mol, Chem.Mol):
                continue

            mol = self.preprocess(mol)
            if mol is None:
                continue

            mols.append(mol)
            indices.append(i)

        out = self.invalid_reward * torch.ones(x.graph.batch_size, device=x.graph.device)
        valids = torch.zeros(x.graph.batch_size, device=x.graph.device, dtype=torch.bool)

        rewards, v = self.compute_rewards(mols, indices, **kwargs)
        rewards = rewards.to(out.device)
        v = v.to(valids.device)
        out[indices] = rewards
        valids[indices] = v
        return out, valids
